﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Text
Imports Microsoft.TeamFoundation.WorkItemTracking.Client
Imports Microsoft.Build.Utilities
Imports Microsoft.Build.Framework
Imports System.Collections
Imports System.Globalization
Imports Microsoft.TeamFoundation
Imports Microsoft.TeamFoundation.Client
Imports Microsoft.TeamFoundation.VersionControl.Client

Namespace MsdnMag.Exchange.PushNotificationClient
    ''' <summary>
    ''' Based on Microsoft.TeamFoundation.Build.Tasks.CreateNewWorkItem.
    ''' </summary>
    Public Class CreateNewWorkItem
        Inherits Task
#Region "Constants"

        Private Const DEFAULT_WORKITEM_TYPE As String = "Bug"
        Private Const DEFAULT_REASON As String = "New"

#End Region

#Region "Private Fields"

        Private _description As String
        Private _properties As String
        Private _workItemTeamProject As String
        Private _title As String
        Private _workItemTypeName As String
        Private _tfsUrl As String
        Private _changesetId As Integer
        Private _assignedTo As String
        Private _workItemId As Integer
        Private _numErrors As Integer
        Private _tfs As TfsTeamProjectCollection
        Private _versionControlServer As VersionControlServer
        Private _workItemStore As WorkItemStore
        Private _associatedWorkItems As String
        Private _attachmentPaths() As String

#End Region

#Region "Methods"

        Public Overrides Function Execute() As Boolean
            Dim workItem As New WorkItem(Me.GetWorkItemType(Me.WorkItemTeamProject, Me.WorkItemType))
            workItem.Title = Me.Title
            workItem.History = Me.Description

            Me.UpdateUserSpecifiedFields(workItem, PropertyParser.GetTable(Me.WorkItemFieldValues))

            ' Add a link to the changeset if one is specified:
            If Me.ChangesetId > 0 Then
                AddLinkToChangeset(Me.ChangesetId, workItem)
            End If

            If Not String.IsNullOrWhiteSpace(Me.AssociatedWorkItems) Then
                Me.AddLinkToAssociatedWorkItems(workItem, Me.AssociatedWorkItems)
            End If

            ' Add any attachments specified:
            If Me.AttachmentPaths IsNot Nothing AndAlso Me.AttachmentPaths.Length > 0 Then
                Me.AddAttachmentsToWorkItems(workItem, Me.AttachmentPaths)
            End If

            Me.SaveWorkItem(workItem)

            Log.LogMessage("CreateNewWorkItem: Workitem Created: Id: {0}.", New Object() {workItem.Id})

            Return True
        End Function

        Private Sub AddAttachmentsToWorkItems(ByVal workItem As WorkItem, ByVal attachmentPaths() As String)
            If workItem Is Nothing OrElse Me.AttachmentPaths.Length = 0 Then
                Return
            End If

            For Each attachmentPath In attachmentPaths
                workItem.Attachments.Add(New Attachment(attachmentPath, ""))
            Next attachmentPath
        End Sub

        Private Sub AddLinkToChangeset(ByVal changesetId As Integer, ByVal workItem As WorkItem)
            Dim changeset As Changeset = Me.VersionControlServer.GetChangeset(changesetId)
            Dim link As Link = New ExternalLink(Me.WorkItemStore.RegisteredLinkTypes("Source Code File"), changeset.ArtifactUri.AbsoluteUri)
            workItem.Links.Add(link)
        End Sub

        Private Sub AddLinkToAssociatedWorkItems(ByVal workItem As WorkItem, ByVal associatedWorkItems As String)
            Dim associatedWorkItemsArray() As String = associatedWorkItems.Split(New Char() {","c}, StringSplitOptions.RemoveEmptyEntries)

            If associatedWorkItemsArray IsNot Nothing AndAlso associatedWorkItemsArray.Length > 0 Then
                For Each item In associatedWorkItemsArray
                    Try
                        Dim associatedWorkItem As WorkItem = Me.WorkItemStore.GetWorkItem(Convert.ToInt32(item))
                        Dim link As New RelatedLink(associatedWorkItem.Id)
                        workItem.Links.Add(link)

                        Log.LogMessage(String.Format("CreateNewWorkItem: Added a link to associated work item: {0}.", New Object() {associatedWorkItem.Id}))
                    Catch ex As Exception
                        Log.LogWarning(String.Format("Unable to associate work item: {0}." & vbCrLf & "Error Message: {1}" & vbCrLf & "StackTrace: {2}.", _
                                                     New Object() {item, ex.Message, ex.StackTrace}))
                    End Try
                Next item
            Else
                Log.LogWarning("CreateNewWorkItem: AssociatedWorkItems list is empty.")
                Return
            End If
        End Sub

        Private Function GetWorkItemType(ByVal projectName As String, ByVal type As String) As WorkItemType
            If String.IsNullOrEmpty(type) Then
                type = DEFAULT_WORKITEM_TYPE
                Log.LogWarning(String.Format("CreateNewWorkItem: Using DefaultWorkItemType: {0}.", New Object() {type}))
            End If

            Try
                Dim project As Project = Me.WorkItemStore.Projects(projectName)
                Return project.WorkItemTypes(type)
            Catch exception As Exception
                Throw New Exception(String.Format("Work Item With Field Not Found: WorkItemType: {0} - WorkItemTeamProject: {1}.", type, Me.WorkItemTeamProject), exception)
            End Try
        End Function

        Private Sub SaveWorkItem(ByVal workItem As WorkItem)
            Try
                workItem.Save()
                Me.WorkItemId = workItem.Id
            Catch exception As Exception
                Dim list = workItem.Validate()
                If list.Count > 0 Then
                    Dim builder As New StringBuilder()
                    For i = 0 To list.Count - 1
                        Dim field As Field = CType(list(i), Field)
                        If i <> 0 Then
                            builder.Append(CultureInfo.CurrentCulture.TextInfo.ListSeparator)
                        End If
                        builder.Append(String.Format("Field Error: {0} : {1}.", New Object() {field.Name, field.Value}))
                    Next i
                    Throw New Exception(String.Format("Field Errors Saving New Work Item: {0}.", New Object() {builder.ToString()}), exception)
                End If
                If Not workItem.Project.HasWorkItemWriteRights Then
                    Throw New Exception("No Permission To Write WI.", exception)
                End If
                Throw
            End Try
        End Sub

        Private Sub SetFieldIfExists(ByVal workItem As WorkItem, ByVal fieldRefName As String, ByVal value As String)
            If workItem.Type.FieldDefinitions.Contains(fieldRefName) Then
                Dim field As Field = workItem.Fields(fieldRefName)
                If field.IsEditable AndAlso ((Not field.IsLimitedToAllowedValues) OrElse field.AllowedValues.Contains(value)) Then
                    workItem(fieldRefName) = value
                End If
            End If
        End Sub

        Private Sub UpdateUserSpecifiedFields(ByVal workItem As WorkItem, ByVal propertiesTable As Dictionary(Of String, String))
            Dim flag As Boolean = False
            If propertiesTable IsNot Nothing Then
                For Each pair As KeyValuePair(Of String, String) In propertiesTable
                    Try
                        Dim field As Field = workItem.Fields(pair.Key)
                        field.Value = pair.Value
                        If TFStringComparer.WorkItemFieldReferenceName.Equals(field.ReferenceName, CoreFieldReferenceNames.Reason) Then
                            flag = True
                        End If
                        Continue For
                    Catch exception As Exception
                        Log.LogWarning(String.Format("Unable To Update WI Field: {0} - {1}.", New Object() {pair.Key, exception.Message}))
                        Continue For
                    End Try
                Next pair
            End If
            If Not flag Then
                Me.SetFieldIfExists(workItem, CoreFieldReferenceNames.Reason, DEFAULT_REASON)
            End If
        End Sub

        Private Sub NonFatalError(ByVal sender As Object, ByVal e As ExceptionEventArgs)
            Dim errorMessage As String = Nothing
            If e.Exception IsNot Nothing Then
                errorMessage = e.Exception.Message
            End If
            If e.Failure IsNot Nothing Then
                If e.Failure.Severity = SeverityType.Warning Then
                    Return
                End If
                errorMessage = e.Failure.Message
            End If
            Log.LogError(errorMessage)
            Me._numErrors += 1
        End Sub

#End Region

#Region "Properties"

        Public Property AssignedTo() As String
            Get
                Return Me._assignedTo
            End Get
            Set(ByVal value As String)
                Me._assignedTo = value
            End Set
        End Property

        Public Property ChangesetId() As Integer
            Get
                Return Me._changesetId
            End Get
            Set(ByVal value As Integer)
                Me._changesetId = value
            End Set
        End Property

        <Required()>
        Public Property Description() As String
            Get
                Return Me._description
            End Get
            Set(ByVal value As String)
                Me._description = value
            End Set
        End Property

        <Required()>
        Public Property WorkItemTeamProject() As String
            Get
                Return Me._workItemTeamProject
            End Get
            Set(ByVal value As String)
                Me._workItemTeamProject = value
            End Set
        End Property

        <Required()>
        Public Property Title() As String
            Get
                Return Me._title
            End Get
            Set(ByVal value As String)
                Me._title = value
            End Set
        End Property

        Public Property AssociatedWorkItems() As String
            Get
                Return _associatedWorkItems
            End Get
            Set(ByVal value As String)
                _associatedWorkItems = value
            End Set
        End Property

        Public Property WorkItemFieldValues() As String
            Get
                Return Me._properties
            End Get
            Set(ByVal value As String)
                Me._properties = value
            End Set
        End Property

        Friend ReadOnly Property WorkItemStore() As WorkItemStore
            Get
                If Me._workItemStore Is Nothing Then
                    _workItemStore = CType(Me.Tfs.GetService(GetType(WorkItemStore)), WorkItemStore)
                End If
                Return _workItemStore
            End Get
        End Property

        Public Property WorkItemType() As String
            Get
                Return Me._workItemTypeName
            End Get
            Set(ByVal value As String)
                Me._workItemTypeName = value
            End Set
        End Property

        <Required()>
        Public Property TeamFoundationServerUrl() As String
            Get
                Return Me._tfsUrl
            End Get
            Set(ByVal value As String)
                Me._tfsUrl = value
            End Set
        End Property

        Friend ReadOnly Property Tfs() As TfsTeamProjectCollection
            Get
                If Me._tfs Is Nothing Then
                    Me._tfs = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(New Uri(Me.TeamFoundationServerUrl))
                End If
                Return Me._tfs
            End Get
        End Property

        Friend ReadOnly Property VersionControlServer() As VersionControlServer
            Get
                If Me._versionControlServer Is Nothing Then
                    Me._versionControlServer = CType(Me.Tfs.GetService(GetType(VersionControlServer)), VersionControlServer)
                    AddHandler _versionControlServer.NonFatalError, AddressOf NonFatalError
                End If
                Return Me._versionControlServer
            End Get
        End Property

        Public Property AttachmentPaths() As String()
            Get
                Return _attachmentPaths
            End Get
            Set(ByVal value As String())
                _attachmentPaths = value
            End Set
        End Property

        <Output()>
        Public Property WorkItemId() As Integer
            Get
                Return Me._workItemId
            End Get
            Set(ByVal value As Integer)
                Me._workItemId = value
            End Set
        End Property

#End Region
    End Class
End Namespace
