﻿Imports Microsoft.VisualBasic
Imports System
Imports MsdnMag.Exchange.PushNotificationSubscriber.Core.EWSReference
Imports System.Net
Imports System.IO
Imports System.Security.Cryptography.X509Certificates
Imports System.Net.Security
Imports System.Xml.Serialization
Imports Microsoft.Exchange.WebServices.Data

Public Class PushNotificationSubscriber
#Region "Constants"

    Private Const MAILBOX_EVENTLOG_TEMPLATE As String = "{0}MailboxEventLog-{1}.txt"
    Private Const TIMESTAMP_TEMPLATE As String = "{0}_{1}_{2}-{3}_{4}_{5}_{6}"
    Private Const CREATE_XML_MESSAGE_TEMPLATE As String = "Writing request and response XML to disk..."
    Private Const REQUEST_XML_FILENAME_TEMPLATE As String = "{0}{1}-{2}.xml"
    Private Const RESPONSE_XML_FILENAME_TEMPLATE As String = "{0}{1}-{2}.xml"
    Private Const SUBSCRIBED_MESSAGE_TEMPLATE As String = "Subscribed for Push notifications: {0}"
    Private Const UNSUBSCRIBED_MESSAGE_TEMPLATE As String = "UnSubscribed from Push notifications. {0}"
    Private Const SUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE As String = vbCrLf & "Subscribed for Push notifications." & vbCrLf & "Subscription Id: {0}"
    Private Const UNSUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE As String = "UnSubscribed from Push notifications." & vbCrLf & "Subscription Id: {0}"
    Private Const SUBSCRIBED_FAILURE_MESSAGE_TEMPLATE As String = vbCrLf & "Subscription for Push notifications was not successful."
    Private Const UNSUBSCRIBED_FAILURE_MESSAGE_TEMPLATE As String = vbCrLf & "UnSubscription from Push notifications was not successful."

#End Region

#Region "Constructors"

    Public Sub New()
    End Sub

#End Region

#Region "Public Properties"

    Public Property EwsServiceUrl() As String
    Public Property PushClientWebServiceUrl() As String
    Public Property BaseLogFolder() As String
    Public Property PushNotificationsFrequencyInMinutes() As Integer
    Private privateSubscriptionId As String
    Public Property SubscriptionId() As String
        Get
            Return privateSubscriptionId
        End Get
        Private Set(ByVal value As String)
            privateSubscriptionId = value
        End Set
    End Property
    Public Property UserName() As String
    Public Property Password() As String
    Public Property Domain() As String
    Public Property UseAutodiscoveryForSubscribing() As Boolean
    Public Property AutoDiscoveryEmail() As String
    Public Property UseCurrentlyLoggedInUserCredentials() As Boolean

#End Region

#Region "Public Methods"

    ''' <summary>
    ''' Creates a push notification subscription.
    ''' </summary>
    Public Sub SubscribeForPushNotifications()
        SubscribeUnsubscribe(False, Nothing)
    End Sub

    ''' <summary>
    ''' Unsubscribes an existing push notification subscription.
    ''' </summary>
    ''' <param name="subscriptionId">The Subscription Id of the push notification subscription to be unsubscribed.</param>
    Public Sub UnSubscribeFromPushNotifications(ByVal subscriptionId As String)
        SubscribeUnsubscribe(True, subscriptionId)
    End Sub

#End Region

#Region "Private Methods"

    ''' <summary>
    ''' A proxy method that prepares the binding information and either subscribes to a push notification or unsubscribes from one.
    ''' </summary>
    ''' <param name="unSubscribe">A boolean flag that indicates whether this is a subscription or an unsubscription.</param>
    ''' <param name="subscriptionId">The Subscription Id of an existing subscription when the unSubscribe flag is set to true.
    ''' Set to null when unSubscribe is set to false.
    ''' </param>
    Private Sub SubscribeUnsubscribe(ByVal unSubscribe As Boolean, ByVal subscriptionId As String)
        System.Net.ServicePointManager.ServerCertificateValidationCallback = Function(obj As Object, certificate As X509Certificate, chain As X509Chain, errors As SslPolicyErrors) True

        ' Create the bindings and set the credentials.
        Dim esb As New ExchangeServiceBinding()
        esb.RequestServerVersionValue = New RequestServerVersion()
        esb.RequestServerVersionValue.Version = ExchangeVersionType.Exchange2010
        esb.Url = EwsServiceUrl
        If UseCurrentlyLoggedInUserCredentials Then
            esb.Credentials = CredentialCache.DefaultNetworkCredentials
        Else
            esb.Credentials = New NetworkCredential(UserName, Password, Domain)
        End If

        If UseAutodiscoveryForSubscribing Then
            Dim service As New ExchangeService()
            service.Credentials = New WebCredentials(esb.Credentials)
            service.AutodiscoverUrl(AutoDiscoveryEmail)
            esb.Url = service.Url.ToString()
        End If

        If unSubscribe Then
            Me.Unsubscribe(esb, subscriptionId)
        Else
            Subscribe(esb)
        End If
    End Sub

    ''' <summary>
    ''' Creates a subscription for the ExchangeServiceBinding instance. 
    ''' </summary>
    ''' <param name="esb">An ExchangeServiceBinding instance containing the binding information.</param>
    Private Sub Subscribe(ByVal esb As ExchangeServiceBinding)
        ' Create a new subscription.
        Dim subscribeRequest As New SubscribeType()
        Dim pushSubscription As New PushSubscriptionRequestType()

        ' Subscribe to events in the inbox folder.
        Dim folders(0) As BaseFolderIdType
        Dim folderId As New DistinguishedFolderIdType()
        folderId.Id = DistinguishedFolderIdNameType.inbox
        folders(0) = folderId
        pushSubscription.FolderIds = folders

        ' Subscribe to new mail event only.
        Dim eventTypes(0) As NotificationEventTypeType
        eventTypes(0) = NotificationEventTypeType.NewMailEvent

        pushSubscription.EventTypes = eventTypes

        ' Receive push notifications every n minutes (typically 1 minute).
        pushSubscription.StatusFrequency = PushNotificationsFrequencyInMinutes

        ' Identify the location of the client Web service.
        pushSubscription.URL = PushClientWebServiceUrl

        ' Form the subscribe request.
        subscribeRequest.Item = pushSubscription

        ' Send the subscribe request and get the response.
        Dim subscribeResponse As SubscribeResponseType = esb.Subscribe(subscribeRequest)

        Dim timestamp As String = GetTimestamp()

        ' Check the result.
        If subscribeResponse.ResponseMessages.Items.Length > 0 AndAlso subscribeResponse.ResponseMessages.Items(0).ResponseClass = ResponseClassType.Success Then
            Dim subscribeResponseMessage As SubscribeResponseMessageType = TryCast(subscribeResponse.ResponseMessages.Items(0), SubscribeResponseMessageType)

            SubscriptionId = subscribeResponseMessage.SubscriptionId

            If Not String.IsNullOrEmpty(BaseLogFolder) Then
                Using sw As New StreamWriter(String.Format(MAILBOX_EVENTLOG_TEMPLATE, BaseLogFolder, timestamp))
                    sw.WriteLine(SUBSCRIBED_MESSAGE_TEMPLATE, SubscriptionId)
                End Using
            End If

            Console.WriteLine(SUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE, SubscriptionId)
        Else
            Console.WriteLine(SUBSCRIBED_FAILURE_MESSAGE_TEMPLATE)
        End If

        CreateXmlMessageTextFile(subscribeRequest, subscribeResponse, BaseLogFolder, timestamp)
    End Sub

    ''' <summary>
    ''' Unsubscribes an existing subscription.
    ''' </summary>
    ''' <param name="esb">An ExchangeServiceBinding instance containing the binding information.</param>
    ''' <param name="subscriptionId">The Subscription Id for the subscription to be unsubscribed.</param>
    Private Sub Unsubscribe(ByVal esb As ExchangeServiceBinding, ByVal subscriptionId As String)
        ' Create a new Unsubscription request.
        Dim unSubscribeRequest As New UnsubscribeType()

        ' Form the subscribe request.
        unSubscribeRequest.SubscriptionId = subscriptionId

        ' Send the subscribe request and get the response.
        Dim unSubscribeResponse As UnsubscribeResponseType = esb.Unsubscribe(unSubscribeRequest)

        Dim timestamp As String = GetTimestamp()

        ' Check the result.
        If unSubscribeResponse.ResponseMessages.Items.Length > 0 AndAlso unSubscribeResponse.ResponseMessages.Items(0).ResponseClass = ResponseClassType.Success Then
            Dim subscribeResponseMessage As SubscribeResponseMessageType = TryCast(unSubscribeResponse.ResponseMessages.Items(0), SubscribeResponseMessageType)

            If Not String.IsNullOrEmpty(BaseLogFolder) Then
                Using sw As New StreamWriter(String.Format(MAILBOX_EVENTLOG_TEMPLATE, BaseLogFolder, timestamp))
                    sw.WriteLine(UNSUBSCRIBED_MESSAGE_TEMPLATE, subscribeResponseMessage.SubscriptionId)
                End Using
            End If

            Console.WriteLine(UNSUBSCRIBED_SUCCESS_MESSAGE_TEMPLATE, subscribeResponseMessage.SubscriptionId)
        Else
            Console.WriteLine(UNSUBSCRIBED_FAILURE_MESSAGE_TEMPLATE)
        End If

        CreateXmlMessageTextFile(unSubscribeRequest, unSubscribeResponse, BaseLogFolder, timestamp)
    End Sub

#End Region

#Region "Static Helper Methods"

    ''' <summary>
    ''' Persists the request (BaseRequestType) and response (BaseResponseMessageType) messages to XML files.
    ''' </summary>
    ''' <param name="request">An instance of BaseRequestType.</param>
    ''' <param name="response">An instance of BaseResponseMessageType.</param>
    ''' <param name="baseLogFolder">The folder where the XML files are to be written.</param>
    ''' <param name="timestamp">A formatted timestamp value with which the XML files will be suffixed.</param>
    Private Shared Sub CreateXmlMessageTextFile(ByVal request As BaseRequestType, ByVal response As BaseResponseMessageType, ByVal baseLogFolder As String, ByVal timestamp As String)
        If String.IsNullOrEmpty(baseLogFolder) Then
            Return
        End If

        Console.WriteLine(CREATE_XML_MESSAGE_TEMPLATE)

        ' Create a text file with the request XML.
        Using myReqWriter As New StreamWriter(String.Format(REQUEST_XML_FILENAME_TEMPLATE, baseLogFolder, request.GetType().Name, timestamp))
            Dim mySerializer As New XmlSerializer(request.GetType())
            mySerializer.Serialize(myReqWriter, request)
        End Using

        ' Create a text file with the response XML.
        Using myRespWriter As New StreamWriter(String.Format(RESPONSE_XML_FILENAME_TEMPLATE, baseLogFolder, response.GetType().Name, timestamp))
            Dim mySerializer As New XmlSerializer(response.GetType())
            mySerializer.Serialize(myRespWriter, response)
        End Using
    End Sub

    ''' <summary>
    ''' Generates a timestamp string to be used as the prefix or suffix of filenames that need to be uniquely distinguished.
    ''' </summary>
    ''' <returns>A formatted string containing the current datetime.</returns>
    Public Shared Function GetTimestamp() As String
        Dim currentTime As Date = Date.Now
        Return String.Format(TIMESTAMP_TEMPLATE, currentTime.Year, currentTime.Month, currentTime.Day, currentTime.Hour, currentTime.Minute, currentTime.Second, currentTime.Millisecond)
    End Function

#End Region
End Class
